/**************************
 * Includes
 *
 **************************/

#include <windows.h>
#include "glut.h"
//#include "vec.h"
#include <stdio.h>

#include "gl_scene.h"

#define KEY_ESC  27
#define KEY_INTRO 13
#define KEY_DELETE 8

#define CON_MAX_LINE_LEN 256
#define	CON_MAX_INPUT_LEN 256

static char	inputString[CON_MAX_INPUT_LEN];
static int	inputMaxLen = CON_MAX_INPUT_LEN;	//max len for this input (caller provided)
static char inputMsg[CON_MAX_LINE_LEN];			//msg to print as input request
static char *inputRet = NULL;					//point where return to.
static int input_mode = FALSE;
static char	*cursor = "_";

void (*inputaction)(void) = NULL;//tmp

int CursorFlashTime(void)
{
   static float lastTime;
   static BOOL flashtime = FALSE;
   float		currentTime;

   currentTime = glutGet(GLUT_ELAPSED_TIME) * 0.001f;
   if( currentTime - lastTime > 0.5f ){
	   lastTime = currentTime;
	   flashtime = !flashtime;
   }

   return flashtime;
}


//===================================================
//					INPUT
//===================================================

//strips based on desired len
int StrCopyLen (char *in, char *out, int len)
{
	int	count = 0;

	while (*in && count <= len ){
		*out++ = *in++;
		count++;
	}
	*out = 0;

	return (strlen(in));	//returns resting len
}


//token consists in the word AND the space(s) after it
int Con_StripWord (char *in, char *token)
{
	char *t;
	t = token;

	while (*in && *in != ' ' ){
		*t++ = *in++;
	}
	//*t = 0;

	//include spaces
	while (*in && *in == ' ' ){
		*t++ = *in++;
	}
	*t = 0;

	return (strlen(in));	//returns resting len
}

//strips a line of (aproximated) desired len, breaking at spaces between words
int Con_StripLine (char *input, char *newline, int linelen)
{
	char	*p;
	int		len = 0;
	char	token[CON_MAX_INPUT_LEN];
	int		rest = strlen(input);

	p = input;

	if (linelen > CON_MAX_LINE_LEN)
		linelen = CON_MAX_LINE_LEN;

	memset(newline, 0, sizeof(newline));

	while (*p && rest && len < linelen)
	{
		rest = Con_StripWord(p, token);
		p += strlen(token);
		len += strlen(token);

		strcat(newline, token);
	}

	return rest;	//return resting len
}

void DrawFormattedText(char *header, char *text, float x, float y, BOOL drawCursor)
{
	int		window_width, window_height, string_width;
	int		characterSize;
	int		numlines = 0;
	float	ypos, yline;
	int		linelen;
	int		le;
	char	newline[CON_MAX_LINE_LEN];
	char	*s;

	window_height = glutGet(GLUT_WINDOW_HEIGHT);
	window_width = glutGet(GLUT_WINDOW_WIDTH);
	window_width *= (1.0f - x);	//adjust line size to available space
	window_width -= (24+(window_width*0.1)); //give it some margins
	characterSize = glutBitmapLength(GLUT_BITMAP_HELVETICA_10, "A");//A sample character.Used for line height size only
	yline = (((float)characterSize*2)*100.0f)/(float)window_height;	//text line height

	//find len for line
	if(!text[0]) {
		linelen = 1;
	} else {
		string_width = glutBitmapLength(GLUT_BITMAP_HELVETICA_10, text);
		linelen = (int) ( (float)window_width / ((float)string_width / (float)strlen(text)) );
	}

	//append cursor (cursor is not counted for linelen)
	if(drawCursor) {
		x += 0.02f;
		if(CursorFlashTime())
			strcat(text, cursor);
		else
			strcat(text, " ");
	}

	//add the input Msg if any
	if ( header ) {
		numlines = 1;
		ypos = y - ((numlines)*yline*0.01f);		//downwards
		drawString( header, GLUT_BITMAP_HELVETICA_10, 1.0f, 1.0f, 1.0f, x , ypos );
	}

	//draw by lines
	le = strlen(text);
	s = text;
	while( le )
	{
		le = Con_StripLine (s, newline, linelen);
		s += strlen(newline);	//move to new origin
		numlines++;
		
		ypos = y - ((numlines)*yline*0.01f);		//downwards
		if(drawCursor)
			drawString( ">", GLUT_BITMAP_HELVETICA_10, 1.0f, 1.0f, 1.0f, x - 0.02f, ypos );
		drawString( newline, GLUT_BITMAP_HELVETICA_10, 1.0f, 1.0f, 1.0f, x , ypos );
	}

	//remove cursor
	if(drawCursor)
		text[strlen(text) - 1] = 0;
}


//===================================================
//					INPUT
//===================================================


int InputStringKey(unsigned char key)
{
	char	newkey[1];

	if(!input_mode)
		return FALSE;

	cursor = "_";

	if(!key)
		return TRUE;

	//printf("key:%i\n", (int)key);//jalcheck

	if(key == KEY_ESC) {
		inputString[0] = 0;
		input_mode = FALSE;
		return TRUE;
	}

	if(key == KEY_INTRO)
	{
		if(inputString[0])
			strncpy( inputRet, inputString, inputMaxLen);

		if(inputaction)
			inputaction();
		inputaction = NULL;

		inputRet = NULL;
		inputString[0] = 0;
		inputMsg[0] = 0;
		input_mode = FALSE;
		return TRUE;
	}

	if(key == KEY_DELETE)
	{
		if(!strlen(inputString))
			return TRUE;
		inputString[strlen(inputString)-1] = 0;
		return TRUE;
	}

	if ((int)strlen(inputString) + 2 > inputMaxLen) {
		cursor = "<";
		return TRUE;
	}

	newkey[0] = key;newkey[1] = 0;
	strcat(inputString, newkey);

	return TRUE;
}


void Input( const char *msg, char *inputPoint, int max_len)
{
	if (input_mode == TRUE)
		return;

	if (max_len < 1) {
		printf("Input LEN is too short\n");
		return;
	}

	if (max_len > CON_MAX_INPUT_LEN ) {
		printf("Input LEN is limited by CON_MAX_INPUT_LEN (%i)\n", CON_MAX_INPUT_LEN);
		max_len = CON_MAX_INPUT_LEN;
	}

	inputMaxLen = max_len;	//limit to the size provided by the caller
	strncpy(inputMsg, msg, CON_MAX_LINE_LEN);
	inputRet = inputPoint;

	input_mode = TRUE;
}



void InputAction( const char *msg, char *inputPoint, int max_len, void (*action)(void))
{
	if (input_mode == TRUE)
		return;

	if (max_len < 1) {
		printf("Input LEN is too short\n");
		return;
	}

	if (max_len > CON_MAX_INPUT_LEN ) {
		printf("Input LEN is limited by CON_MAX_INPUT_LEN (%i)\n", CON_MAX_INPUT_LEN);
		max_len = CON_MAX_INPUT_LEN;
	}

	inputMaxLen = max_len;	//limit to the size provided by the caller
	strncpy(inputMsg, msg, CON_MAX_LINE_LEN);
	inputRet = inputPoint;

	inputaction = action;

	input_mode = TRUE;
}







void AddInputWindow(void)
{
	if(!input_mode)
		return;

	DrawFormattedText( inputMsg, inputString, 0.02f, 1.0f, TRUE );
}

//JALFIXME

#ifdef _WIN32
# ifndef vsnprintf 
#  define vsnprintf _vsnprintf
# endif
#endif

void SCRPrint (float x, float y, char *fmt, ...)
{
	va_list		argptr;
	//char		msg[MAXPRINTMSG];
	char		msg[256];

	//if (!developer || !developer->integer)
	//	return;			// don't confuse non-developers with techie stuff...

	va_start (argptr,fmt);
	vsnprintf (msg, sizeof(msg), fmt, argptr);
	//_vsnprintf (msg, sizeof(msg), fmt, argptr);
	va_end (argptr);

	//Com_Printf ("%s", msg);

	DrawFormattedText( NULL, msg, x, y, FALSE );
}
