/**************************
 * Includes
 *
 **************************/

#include <windows.h>
#include "glut.h"
//#include "vec.h"
#include <stdio.h>

#include "gl_scene.h"


int wireframe;
int show_fps = 1;
extern int show_quats;



//--------------------------------------------------

/* textures */
#define MAX_TEXTURES 	 1  	// max textures displayed

typedef struct												// Create A Structure
{
	GLubyte	*imageData;										// Image Data (Up To 32 Bits)
	GLuint	bpp;											// Image Color Depth In Bits Per Pixel.
	GLuint	width;											// Image Width
	GLuint	height;											// Image Height
	GLuint	texID;											// Texture ID Used To Select A Texture
} TextureImage;												// Structure Name

TextureImage textures[MAX_TEXTURES];	// array to hold texture handles


BOOL LoadTGATexture(TextureImage *texture, char *filename)			// Loads A TGA File Into Memory
{    
	GLubyte		TGAheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};	// Uncompressed TGA Header
	GLubyte		TGAcompare[12];								// Used To Compare TGA Header
	GLubyte		header[6];									// First 6 Useful Bytes From The Header
	GLuint		bytesPerPixel;								// Holds Number Of Bytes Per Pixel Used In The TGA File
	GLuint		imageSize;									// Used To Store The Image Size When Setting Aside Ram
	GLuint		temp;										// Temporary Variable
	GLuint		type=GL_RGBA;								// Set The Default GL Mode To RBGA (32 BPP)
	GLuint i;

	FILE *file = fopen(filename, "rb");						// Open The TGA File

	if(	file==NULL ||										// Does File Even Exist?
		fread(TGAcompare,1,sizeof(TGAcompare),file)!=sizeof(TGAcompare) ||	// Are There 12 Bytes To Read?
		memcmp(TGAheader,TGAcompare,sizeof(TGAheader))!=0				||	// Does The Header Match What We Want?
		fread(header,1,sizeof(header),file)!=sizeof(header))				// If So Read Next 6 Header Bytes
	{
		if (file == NULL)									// Did The File Even Exist? *Added Jim Strong*
			return FALSE;									// Return False
		else
		{
			fclose(file);									// If Anything Failed, Close The File
			return FALSE;									// Return False
		}
	}

	texture->width  = header[1] * 256 + header[0];			// Determine The TGA Width	(highbyte*256+lowbyte)
	texture->height = header[3] * 256 + header[2];			// Determine The TGA Height	(highbyte*256+lowbyte)
    
 	if(	texture->width	<=0	||								// Is The Width Less Than Or Equal To Zero
		texture->height	<=0	||								// Is The Height Less Than Or Equal To Zero
		(header[4]!=24 && header[4]!=32))					// Is The TGA 24 or 32 Bit?
	{
		fclose(file);										// If Anything Failed, Close The File
		return FALSE;										// Return False
	}

	texture->bpp	= header[4];							// Grab The TGA's Bits Per Pixel (24 or 32)
	bytesPerPixel	= texture->bpp/8;						// Divide By 8 To Get The Bytes Per Pixel
	imageSize		= texture->width*texture->height*bytesPerPixel;	// Calculate The Memory Required For The TGA Data

	texture->imageData=(GLubyte *)malloc(imageSize);		// Reserve Memory To Hold The TGA Data

	if(	texture->imageData==NULL ||							// Does The Storage Memory Exist?
		fread(texture->imageData, 1, imageSize, file)!=imageSize)	// Does The Image Size Match The Memory Reserved?
	{
		if(texture->imageData!=NULL)						// Was Image Data Loaded
			free(texture->imageData);						// If So, Release The Image Data

		fclose(file);										// Close The File
		return FALSE;										// Return False
	}

	for(i=0; i<(int)imageSize; i+=bytesPerPixel)		// Loop Through The Image Data
	{														// Swaps The 1st And 3rd Bytes ('R'ed and 'B'lue)
		temp=texture->imageData[i];							// Temporarily Store The Value At Image Data 'i'
		texture->imageData[i] = texture->imageData[i + 2];	// Set The 1st Byte To The Value Of The 3rd Byte
		texture->imageData[i + 2] = temp;					// Set The 3rd Byte To The Value In 'temp' (1st Byte Value)
	}

	fclose (file);											// Close The File

	// Build A Texture From The Data
	glGenTextures(1, &texture[0].texID);					// Generate OpenGL texture IDs

	glBindTexture(GL_TEXTURE_2D, texture[0].texID);			// Bind Our Texture
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	// Linear Filtered
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	// Linear Filtered
	
	if (texture[0].bpp==24)									// Was The TGA 24 Bits
	{
		type=GL_RGB;										// If So Set The 'type' To GL_RGB
	}

	glTexImage2D(GL_TEXTURE_2D, 0, type, texture[0].width, texture[0].height, 0, type, GL_UNSIGNED_BYTE, texture[0].imageData);

	return TRUE;											// Texture Building Went Ok, Return True
}


void InitializeOGLTextures(GLvoid)											// All Setup For OpenGL Goes Here
{
	glEnable(GL_TEXTURE_2D);
	if (!LoadTGATexture(&textures[0],"logo.tga"))				// Load The Font Texture
		return;										// If Loading Failed, Return False

	glBindTexture(GL_TEXTURE_2D, textures[0].texID);
}

//-------------------------------------------------------

///vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
/// The code here can be freely altered too. It's mainly designed to set
/// OpenGl's initial value and configuration.
///vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
void InitializeOGL(void)
{
	glEnable(GL_TEXTURE_2D);						// Enable Texture Mapping 
	glShadeModel(GL_SMOOTH);						// Enables Smooth Shading
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);					// Black Background

	glClearDepth(1.0f);								// Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);						// Enables Depth Testing
	glDepthFunc(GL_LEQUAL);							// The Type Of Depth Test To Do

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);			// Really Nice Perspective Calculations

	InitializeOGLTextures();
}

//-------------------------------------------------------


//dirty copy
void Str_StripExtension (char *in, char *out)
{
	while (*in && *in != '.')
		*out++ = *in++;
	*out = 0;
}

void SKM_ReadSKMfile(char* filename);
//-------------------------------------------------------
//
//		SCENE
//
//-------------------------------------------------------
void Init_3dObjects(void)
{
//	object_t	*obj;

	printf("Init 3d Objects\n");
	//Init sceneObjects
	memset (&sceneObjects, 0, sizeof(sceneObjects));
	sceneNumObjects = 0;

	//-------------tmp SKM load
	SKM_ReadSKMfile("data/tris.skm");
}


void Init_GLCam(void)
{
	cam.flyMode = 0;

	//fly mode
    cam.origin[X] = 0;
    cam.origin[Y] = 0;
    cam.origin[Z] = 128.0f;

	cam.angles[YAW] = 0;
    cam.angles[PITCH] = 0;
    cam.angles[ROLL] = 0;

	//rotation mode
	cam.eye[0] = 0.0f;
	cam.eye[1] = 0.0f;
	cam.eye[2] = 128.0f;

	cam.rot[0] = 0.0f;
	cam.rot[1] = 0.0f;
	cam.rot[2] = 0.0f;
}

void InitializeWorld (void)
{
	Init_GLCam();
	Init_3dObjects();
}


void CAM_ToogleFlyMode(void)
{
	cam.flyMode = !cam.flyMode;
}

void DISPLAY_ToogleShowFPS(void)
{
	show_fps = !show_fps;
}

void DISPLAY_ToogleWireframe(void)
{
	wireframe = !wireframe;
}

void DISPLAY_ToogleQuats(void)
{
	show_quats = !show_quats;
}

void DISPLAY_ToogleFullScreen(int width, int height)
{
}

//------------------------------------------------------------------------
// Draws a string on screen
//------------------------------------------------------------------------
void drawString(char *str, void *font, GLclampf r, GLclampf g, GLclampf b,
              GLfloat x, GLfloat y) {
  // font: font to use, e.g., GLUT_BITMAP_HELVETICA_10
  //   r, g, b: text colour
  //   x, y: text position in window: range [0,0] (bottom left of window)
  //         to [1,1] (top right of window). 
	
	char *ch;
	GLint matrixMode;
	GLboolean lightingOn;
	
	lightingOn= glIsEnabled(GL_LIGHTING);        // lighting on? 
	if (lightingOn) glDisable(GL_LIGHTING);
	
	glGetIntegerv(GL_MATRIX_MODE, &matrixMode);  // matrix mode? 
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glPushAttrib(GL_COLOR_BUFFER_BIT);       // save current colour
	glColor3f(r, g, b);
	glRasterPos3f(x, y, 0.0);
	for(ch= str; *ch; ch++) {
		glutBitmapCharacter(font, (int)*ch);
	}
	glPopAttrib();
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(matrixMode);
	if (lightingOn) glEnable(GL_LIGHTING);
}

//------------------------------------------------------------------------
// Add2DText: Add text strings
//------------------------------------------------------------------------
static char	FPSstring[50];
void FPScalculator(int time, int framecount)
{
	static int lastFPStime;
	static int lastFPSframecount;
	int	FPScount;

	if(time - lastFPStime >= 1000)
	{
		lastFPStime = time;
		FPScount = framecount - lastFPSframecount;
		lastFPSframecount = framecount;

		//update the string
		sprintf( FPSstring, "FPS: %i", FPScount );
	}
}
void DrawFPS(void)
{
	if(!show_fps)
		return;

    drawString(FPSstring, GLUT_BITMAP_HELVETICA_10, 1.0f, 1.0f, 1.0f, 0.02f, 0.12f );
}

void SKM_DrawModelInfo(void);
void AddText2D(void)
{
	DrawFPS();
	SKM_DrawModelInfo();
}



//-------------------------------------------------------------
// Objects drawing
//-------------------------------------------------------------

//origin + object coords (don't translate)
int GL_AddLineToScene(vec3_t origin, vec3_t origin2, vec3_t color)                                        // Here's Where We Do All The Drawing
{
	glPushMatrix();
    glColor3f(color[RED],color[GREEN],color[BLUE]);
	//triangle
	glBegin(GL_LINES);
		glVertex3f( origin[1] , origin[2], origin[0]  );
		glVertex3f( origin2[1] , origin2[2] , origin2[0] );
	glEnd();
	glPopMatrix();

	return TRUE;
}


int GL_AddCubeToSceneAtLoc(vec3_t origin, vec3_t color, vec3_t angles, float scale)
{
    GLfloat size;
    GLfloat x, y, z;
	z = 0.0f;
    y = 0.0f;
    x = 0.0f;
	
	glPushMatrix();
	glTranslatef ( origin[1], origin[2], origin[0]);
	glRotatef( angles[YAW], 0.0f, 1.0f, 0.0f );//ROLL,YAW,PITCH
	glRotatef( angles[PITCH], 0.0f, 0.0f, 1.0f );
	glRotatef( angles[ROLL], 1.0f, 0.0f, 0.0f );

	glColor3f(color[RED],color[GREEN],color[BLUE]);
	size = scale;
	
	glBegin(GL_QUADS);
	    //top
		glVertex3f( x+size, y+size, z-size);			// Top Right Of The Quad (Top)
		glVertex3f( x-size, y+size, z-size);			// Top Left Of The Quad (Top)
		glVertex3f( x-size, y+size, z+size);			// Bottom Left Of The Quad (Top)
		glVertex3f( x+size, y+size, z+size);			// Bottom Right Of The Quad (Top)
		
		//bottom
		glVertex3f( x+size, y-size, z+size);			// Top Right Of The Quad (Bottom)
		glVertex3f( x-size, y-size, z+size);			// Top Left Of The Quad (Bottom)
		glVertex3f( x-size, y-size, z-size);			// Bottom Left Of The Quad (Bottom)
		glVertex3f( x+size, y-size, z-size);			// Bottom Right Of The Quad (Bottom)
		
		//front
		glVertex3f( x+size, y+size, z+size);			// Top Right Of The Quad (Front)
		glVertex3f( x-size, y+size, z+size);			// Top Left Of The Quad (Front)
		glVertex3f( x-size, y-size, z+size);			// Bottom Left Of The Quad (Front)
		glVertex3f( x+size, y-size, z+size);			// Bottom Right Of The Quad (Front)
		
		//back
		glVertex3f( x+size, y-size, z-size);			// Bottom Left Of The Quad (Back)
		glVertex3f( x-size, y-size, z-size);			// Bottom Right Of The Quad (Back)
		glVertex3f( x-size, y+size, z-size);			// Top Right Of The Quad (Back)
		glVertex3f( x+size, y+size, z-size);			// Top Left Of The Quad (Back)
		
		//left
		glVertex3f( x-size, y+size, z+size);			// Top Right Of The Quad (Left)
		glVertex3f( x-size, y+size, z-size);			// Top Left Of The Quad (Left)
		glVertex3f( x-size, y-size, z-size);			// Bottom Left Of The Quad (Left)
		glVertex3f( x-size, y-size, z+size);			// Bottom Right Of The Quad (Left)
		
		//right
		glVertex3f( x+size, y+size, z-size);			// Top Right Of The Quad (Right)
		glVertex3f( x+size, y+size, z+size);			// Top Left Of The Quad (Right)
		glVertex3f( x+size, y-size, z+size);			// Bottom Left Of The Quad (Right)
		glVertex3f( x+size, y-size, z-size);			// Bottom Right Of The Quad (Right)
	glEnd();
	glPopMatrix();
	
	return TRUE;
}

/*
void GL_AddObjectsToScene()
{
	int		i;
	object_t *obj;
	
	for( i=0, obj = sceneObjects; i<sceneNumObjects; i++, obj++) {
		
		switch (obj->modelType)
		{
		default:
		case MODEL_NONE:
			break;
		}
	}
}
*/

//------------------------------------------------------------------------
// If rotation angle is greater of 360 or lesser than -360,
// resets it back to zero.//THIS IS JUST WRONG
//------------------------------------------------------------------------
void clampAngle(float a)
{
	if (a > 360)
		a -= 360;
	if ( a < -360 )
		a += 360;
}
//---------------------
//Cammera Movements
//---------------------
void CamMoveUp(float value)
{
	cam.origin[Y] += (float)0.05f * value;
}
void CamMoveDown(float value)
{
	cam.origin[Y] -= (float)0.05f * value;
}
void CamMoveFwrd(float value)
{
	cam.origin[X] -= (float)sin(cam.angles[YAW]*PIOVER180) * 5.0f * value;
	cam.origin[Z] -= (float)cos(cam.angles[YAW]*PIOVER180) * 5.0f * value;
	cam.origin[Y] -= (float)sin(cam.angles[PITCH]*PIOVER180) * 5.0f * value;
}
void CamMoveBack(float value)
{
	cam.origin[X] += (float)sin(cam.angles[YAW]*PIOVER180) * 5.0f * value;
	cam.origin[Z] += (float)cos(cam.angles[YAW]*PIOVER180) * 5.0f * value;
	cam.origin[Y] += (float)sin(cam.angles[PITCH]*PIOVER180) * 5.0f * value;
}
void CamMoveLeft(float value)
{
	float tmpYaw;
	tmpYaw = cam.angles[YAW] - 90;

	cam.origin[X] += (float)sin(tmpYaw*PIOVER180) * 5.0f * value;
	cam.origin[Z] += (float)cos(tmpYaw*PIOVER180) * 5.0f * value;
}
void CamMoveRight(float value)
{
	float tmpYaw;
	tmpYaw = cam.angles[YAW] + 90;

	cam.origin[X] += (float)sin(tmpYaw*PIOVER180) * 5.0f * value;
	cam.origin[Z] += (float)cos(tmpYaw*PIOVER180) * 5.0f * value;
}
void CamRotateYaw(float x)
{
	cam.angles[YAW] += x;
	clampAngle(cam.angles[YAW]);
}
void CamRotatePitch(float y)
{
	cam.angles[PITCH] += y;
	clampAngle(cam.angles[PITCH]);
}


//ROTATION MODE
//===============================================================
void clamp(float *v)
{
    int i;
	
    for (i = 0; i < 3; i ++)
        if (v[i] > 360 || v[i] < -360)
            v[i] = 0;
}
void CamRotateScreen( float x, float y)
{
	// rotates screen
	cam.rot[0] += y;
	cam.rot[1] += x;
	clamp (cam.rot);
}
void CamTranslateZ(float value)
{
	cam.eye[2] += value; // 0.2 is reduction factor
	clamp (cam.rot);
}
void CamTranslateXY(float x, float y)
{
	cam.eye[0] += x;
	cam.eye[1] += y;
	clamp (cam.rot);
}
//===============================================================
//ROTATION MODE


//-------------------------------------------------------------
// DrawGLScene
//-------------------------------------------------------------
void SKMDrawModel(void);//tmp
void DrawImputString(void);//tmp
void AddInputWindow(void);
void DrawGLScene(void)                                        // Here's Where We Do All The Drawing
{
    //cam data[start]
	GLfloat xtrans;
	GLfloat ztrans;
	GLfloat ytrans;
	GLfloat sceneroty;//cam [end]

	if (wireframe)
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    else
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	
	//Clear for New frame
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);        // Clear The Screen And The Depth Buffer
	glPushMatrix();
	

	if(cam.flyMode)
	{
		glLoadIdentity();   // Reset The Current Modelview Matrix

		//cam position setup [start]
		xtrans = -cam.origin[X];
		ytrans = -cam.origin[Y];
		ztrans = -cam.origin[Z];
		sceneroty = 360.0f - cam.angles[YAW];
		glRotatef(cam.angles[PITCH],1.0f,0,0);
		glRotatef(sceneroty,0,1.0f,0);
		glTranslatef(xtrans, ytrans, ztrans);
		//cam set up [end]

	} else {
		
		glTranslatef (-cam.eye[0], -cam.eye[1], -cam.eye[2]);	// translate camera to the coordinates provided by eye array
		glRotatef(cam.rot[0], 1.0f, 0.0f, 0.0f);	// rotates the screen by the angles provided by rot array
		glRotatef(cam.rot[1], 0.0f, 1.0f, 0.0f);
		glRotatef(cam.rot[2], 0.0f, 0.0f, 1.0f);
	}

	SKMDrawModel();	//tmp (animation blending)
	//GL_AddObjectsToScene();
	AddInputWindow();
	AddText2D();

	glPopMatrix();
	//glFlush();
    glutSwapBuffers();
	glutPostRedisplay();
}





void KEYS_CheckKeyStates(void);
void Scene_Main( void )
{
#ifdef WIN32
	timer.time = timeGetTime();
#else
	timer.time = glutGet(GLUT_ELAPSED_TIME);
#endif
	
	timer.frametime = (int)(1000.0f / MAX_FPS);	//move me to init

	if( timer.time - timer.lasttime < timer.frametime )
		return;

	FPScalculator( timer.time, timer.framecount );

	timer.lasttime = timer.time;
	timer.framecount++;

	KEYS_CheckKeyStates();

	DrawGLScene();
}



