

#include <windows.h>
#include "glut.h"
#include "vec.h"
#include <stdio.h>

#include "gl_scene.h"

#include "endian.h"
#include "skm_model.h"




//=============================================================
//=============================================================
void writefile (char *filename, void *buffer, int size)
{
	int size1;
	FILE *file;

	file = fopen (filename, "wb");
	if (!file)
	{
		printf ("unable to open file \"%s\" for writing\n", filename);
		return;
	}

	size1 = fwrite (buffer, 1, size, file);
	fclose (file);

	if (size1 < size)
	{
		printf ("unable to write file \"%s\"\n", filename);
		return;
	}
}

unsigned char *output;
unsigned char outputbuffer[SKM_MAX_FILESIZE];

void putstring (char *in, int length)
{
	while (*in && length)
	{
		*output++ = *in++;
		length--;
	}

	// pad with nulls
	while (length--)
		*output++ = 0;
}

void putnulls (int num)
{
	while (num--)
		*output++ = 0;
}

void putlong (int num)
{
	*output++ = ((num >>  0) & 0xFF);
	*output++ = ((num >>  8) & 0xFF);
	*output++ = ((num >> 16) & 0xFF);
	*output++ = ((num >> 24) & 0xFF);
}

void putfloat (double num)
{
	union
	{
		float f;
		int i;
	}
	n;

	n.f = (float)num;

	// this matches for both positive and negative 0, thus setting it to positive 0
	if (n.f == 0)
		n.f = 0;

	putlong (n.i);
}

void putinit (void)
{
	output = outputbuffer;
}

int putgetposition (void)
{
	return (int) output - (int) outputbuffer;
}

void putsetposition (int n)
{
	output = (unsigned char *)(n + (int) outputbuffer);
}
//=============================================================
//=============================================================



//---------------------------------------

//tmp holder
mskmodel_t *modelSKM;

void SKM_SetNewFrame(mskmodel_t *model);
void SKM_InitBoneKeys(mskmodel_t *model);

//---------------------------------------



//----------------------------------------------------------
void SKM_saveSKPfile(mskmodel_t *model)
{
	int		i, j, k;
	//---------------------
	int		pos_filesize;
	int		pos_offsets;
	int		pos_bones;
	int		pos_frames;
	int		pos_framebones;
	//---------------------
	int		restoreposition;
	int		filesize;
	//---------------------
	
	// create filehandle
	putinit ();

	// write header
	//======================

	// ID string
	putstring (SKMHEADER, strlen(SKMHEADER));
	
	// model type
	putlong (SKM_MODELTYPE);
	
	// filesize
	pos_filesize = putgetposition ();
	putlong (0);
	
	putlong (model->numbones);
	putlong (model->numframes);
	
	// offsets to things
	pos_offsets = putgetposition ();
	putlong (0);
	putlong (0);


	// write bones
	//======================
	pos_bones = putgetposition ();

	for (i = 0; i < (int)model->numbones; i++)
	{
		putstring (model->bones[i].name, SKM_MAX_NAME);
		putlong (model->bones[i].parent);
		putlong (model->bones[i].flags);
	}
	
	// write frames
	//======================
	pos_frames = putgetposition ();

	// skip over the frame structs, they will be filled in later
	putsetposition (pos_frames + model->numframes * (SKM_MAX_NAME + 4));
	
	for (i = 0; i < (int)model->numframes; i++)
	{
		pos_framebones = putgetposition ();
		for (j = 0; j < (int)model->numbones; j++)
		{
			for (k = 0; k < 4; k++)
				putfloat (model->frames[i].boneposes[j].quat[k]);
			for (k = 0; k < 3; k++)
				putfloat (model->frames[i].boneposes[j].origin[k]);
		}
		
		// now we actually write the frame header
		restoreposition = putgetposition ();
		putsetposition (pos_frames + i * (SKM_MAX_NAME + 4));
		putstring (model->frames[i].name, SKM_MAX_NAME);
		putlong (pos_framebones);
		putsetposition (restoreposition);
	}


	// Finish: fill header data
	//==========================
	filesize = putgetposition ();
	putsetposition (pos_offsets);
	putlong (pos_bones);
	putlong (pos_frames);
	putsetposition (pos_filesize);
	putlong (filesize);
	putsetposition (filesize);

	// write the file
	printf ("file size: %5ik\n", (filesize + 1023) >> 10);
	writefile ("data/out.skp", outputbuffer, filesize);
	printf ("wrote SKP file\n");
}


void SKM_saveSKMfile(mskmodel_t *model)
{
	int i, j;
	int	k, l;
	//-------------------------
	int pos_filesize;
	int	pos_ofs_meshes;
	int	pos_meshes;
	int	pos_verts;
	int	pos_texcoords;
	int	pos_index;
	int	pos_references;
	//-------------------------
	int filesize;
	int restoreposition;
	//-------------------------
	mskvertex_t		*poutskmvert;
	mskbonevert_t	*poutbonevert;
	vec2_t			*poutstcoord;
	index_t			*pouttris;
	unsigned int	*poutreferences;


	// create filehandle
	putinit ();

	// write header
	//======================

	// ID string
	putstring (SKMHEADER, strlen(SKMHEADER));

	// model type
	putlong (SKM_MODELTYPE);

	// filesize
	pos_filesize = putgetposition ();
	putlong (0);

	// numbers of things
	putlong (model->numbones);
	putlong (model->nummeshes);

	// offsets to things
	pos_ofs_meshes = putgetposition ();
	putlong (0);

	// store the meshes
	pos_meshes = putgetposition ();

	// skip over the mesh structs, they will be filled in later
	putsetposition (pos_meshes + model->nummeshes * (SKM_MAX_NAME*2 + 28));

	//mesh cycle
	for (i = 0; i < (int)model->nummeshes; i++)
	{
		//write vertices (ofs_verts)
		//================================
		pos_verts = putgetposition ();

		poutskmvert = model->meshes[i].vertexes;
		for( j = 0; j < (int)model->meshes[i].numverts; j++, poutskmvert++ ) 
		{
			putlong(poutskmvert->numbones);
			poutbonevert = model->meshes[i].vertexes[j].verts;
		
			for( l = 0; l < (int)poutskmvert->numbones; l++, poutbonevert++ ) 
			{
				for( k = 0; k < 3; k++ )
					putfloat(poutbonevert->origin[k]);
				
				putfloat(poutbonevert->influence);

				for( k = 0; k < 3; k++ )
					putfloat(poutbonevert->normal[k]);
				
				putlong(poutbonevert->bonenum);
			}
		}

		//write stcoords (ofs_stcoords)
		//================================
		pos_texcoords = putgetposition ();

		poutstcoord = model->meshes[i].stcoords;

		for( j = 0; j < (int)model->meshes[i].numverts; j++ ) {
			putfloat(poutstcoord[j][0]);
			putfloat(poutstcoord[j][1]);
		}

		//write indexes (ofs_indices)
		//================================
		pos_index = putgetposition ();

		pouttris = model->meshes[i].indexes;

		for( j = 0; j < (int)model->meshes[i].numtris; j++, pouttris += 3 ) {
			putlong(pouttris[0]);
			putlong(pouttris[1]);
			putlong(pouttris[2]);
		}

		//write references (ofs_references)
		//================================
		pos_references = putgetposition ();

		poutreferences = model->meshes[i].references;

		for( j = 0; j < (int)model->meshes[i].numreferences; j++, poutreferences++ )
			putlong(*poutreferences);


		//write mesh header
		//================================
		restoreposition = putgetposition ();
		putsetposition (pos_meshes + i * (SKM_MAX_NAME*2 + 28));
		putstring (model->meshes[i].shadername, SKM_MAX_NAME);
		putstring (model->meshes[i].name, SKM_MAX_NAME);

		putlong (model->meshes[i].numverts);
		putlong (model->meshes[i].numtris);
		putlong (model->meshes[i].numreferences);
		putlong (pos_verts);
		putlong (pos_texcoords);
		putlong (pos_index);
		putlong (pos_references);
		putsetposition (restoreposition);

	}

	//finish: SKM file header
	filesize = putgetposition ();
	putsetposition (pos_ofs_meshes);
	putlong (pos_meshes);
	putsetposition (pos_filesize);
	putlong (filesize);
	putsetposition (filesize);	//restore position

	printf ("file size: %5ik\n", (filesize + 1023) >> 10);
	writefile ("data/out.skm", outputbuffer, filesize);
	printf ("wrote SKM file\n");
}


//----------------------------------------------------------

void SKM_SaveSkeleton(void)
{
	SKM_saveSKPfile(modelSKM);
}

void SKM_SaveModel(void)
{
	SKM_saveSKMfile(modelSKM);
}

//-------------------------------------------

//-------------------------------------------
//		ANIMATIONS
//-------------------------------------------

//-------------------------------------------

//-------------------------------------------
enum {
	LOWER = 0,	
	UPPER,
	HEAD,

	PMODEL_PARTS
};

enum {
	BASIC_CHANNEL,	
	EVENT_CHANNEL,

	ANIMBUFFER_CHANNELS
};

typedef struct
{
	int		newanim[PMODEL_PARTS];
	
} animationbuffer_t;



#define MAX_MODEL_ANIMS 1024
typedef struct
{
	int		firstframe;		//animation script
	int		lastframe;
	int		loopingframes;

}animation_t;

typedef struct animationconfig_s
{
	int		numAnims;
	float	frametime;
	animation_t		anims[MAX_MODEL_ANIMS];

}animationconfig_t;

typedef struct
{
	int		current[PMODEL_PARTS];			//running animation
	int		currentChannel[PMODEL_PARTS];	//what channel it came from

	int		frame[PMODEL_PARTS];
	int		oldframe[PMODEL_PARTS];

	//splitskel
	float	prevframetime;	//interpolation timing stuff
	float	nextframetime;
	float	backlerp;

	animationbuffer_t	buffer[ANIMBUFFER_CHANNELS];

	//config data
	animationconfig_t	config;

} animationinfo_t;

animationinfo_t skmanim;


BOOL LoadAnimationConfig(mskmodel_t *model)
{
	return FALSE;
}

void CreateAnimationConfig(mskmodel_t *model)
{
	skmanim.config.anims[skmanim.config.numAnims].firstframe = 1;
	skmanim.config.anims[skmanim.config.numAnims].lastframe = model->numframes;
	skmanim.config.anims[skmanim.config.numAnims].loopingframes = model->numframes;

	skmanim.config.numAnims = 1;
}

//-------------------------------------------

//-------------------------------------------
//	end animations
//-------------------------------------------

//-------------------------------------------

//--------------------------------------------


//----------------------------------------------------------

void SKM_ReadSKPfile(char* filename, mskmodel_t	*poutmodel)
{
	FILE*		file;
	unsigned char	*buffer;

	int			filesize, pos=0;
	int			i, j, k;
	//----------------------------
	dskpheader_t	*pinmodel;
	dskpbone_t		*pinbone;
	dskpframe_t		*pinframe;
	dskpbonepose_t	*pinbonepose;
	//----------------------------
	mskbone_t		*poutbone;
	mskframe_t		*poutframe;
	mskbonepose_t	*poutbonepose;
	//----------------------------
	
	// open the file to get the size
	file = fopen( filename, "rb");
	if (!file) {
		printf("ReadSKP failed: can't open data file \"%s\".\n",filename);
		exit(1);
	}
	// get filesize
	pos = ftell( file );
	fseek( file, 0, SEEK_END );
	filesize = ftell( file );
	fseek( file, pos, SEEK_SET );
	fclose( file );


	// Having the size, read it
	//-------------------------------------------------------
	buffer = malloc( filesize * sizeof(unsigned char) );
	file = fopen(filename, "rb");
	fread( buffer, filesize ,sizeof(unsigned char), file);
	fclose(file);
	
	if(!buffer){
		printf("ReadSKP failed: Empty data file \"%s\".\n",filename);
		exit(1);
	}
	printf("Reading SKP model %s\n", filename);
	//-------------------------------------------------------


	pinmodel = ( dskpheader_t * )buffer;
	printf ( "----------------------\n");
	printf ( "%s pinmodel type: %i\n", filename, pinmodel->type);
	printf ( "%s pinmodel filesize: %i (MAX: %i)\n", filename, pinmodel->filesize, SKM_MAX_FILESIZE);
	printf ( "%s pinmodel numbones: %i\n", filename, pinmodel->num_bones);
	if( LittleLong (pinmodel->type) != SKM_MODELTYPE )
		printf ( "%s has wrong type number (%i should be %i)\n",
		filename, LittleLong (pinmodel->type), SKM_MODELTYPE );
	if( LittleLong (pinmodel->type) != pinmodel->type || 
		LittleLong (pinmodel->type) > SKM_MAX_BONES) {
		printf ( "%s num_bones: Invalid value\n", filename );
	}

	poutmodel->numframes = LittleLong ( pinmodel->num_frames );
	printf("numframes:pout:%i pin:%i\n", poutmodel->numframes, LittleLong ( pinmodel->num_frames ));
	if( poutmodel->numframes <= 0 )
		printf("%s has no frames", filename );
	else if ( poutmodel->numframes > SKM_MAX_FRAMES )
		printf("%s has too many frames", filename );

	//-----------
	//copy numbones (at QF it is set up at loading the model)

	poutmodel->numbones = LittleLong ( pinmodel->num_bones );
	printf("NumBones(SKP): %i\n",(int)poutmodel->numbones);

	pinbone = ( dskpbone_t * )( ( unsigned char * )pinmodel + LittleLong ( pinmodel->ofs_bones ) );
	poutbone = poutmodel->bones = malloc ( sizeof(mskbone_t) * poutmodel->numbones );

	for( i = 0; i < (int)poutmodel->numbones; i++, pinbone++, poutbone++ ) 
	{
		strncpy( poutbone->name, pinbone->name, SKM_MAX_NAME );
		poutbone->flags = LittleLong ( pinbone->flags );
		poutbone->parent = LittleLong ( pinbone->parent );
		//printf("bone: %s - %s\n",pinbone->name, poutbone->name);
	}

	pinframe = ( dskpframe_t * )( ( unsigned char * )pinmodel + LittleLong ( pinmodel->ofs_frames ) );
	poutframe = poutmodel->frames = malloc ( sizeof(mskframe_t) * poutmodel->numframes );

	for( i = 0; i < (int)poutmodel->numframes; i++, pinframe++, poutframe++ ) {
		poutbone = poutmodel->bones;

		//jal: frames names are ignored at QFusion. I added them here for saving changes
		strncpy( poutframe->name, pinframe->name, SKM_MAX_NAME );

		pinbonepose = ( dskpbonepose_t * )( ( unsigned char * )pinmodel + LittleLong (pinframe->ofs_bonepositions) );
		poutbonepose = poutframe->boneposes = malloc ( sizeof(mskbonepose_t) * poutmodel->numbones );

		for( j = 0; j < (int)poutmodel->numbones; j++, poutbone++, pinbonepose++, poutbonepose++ ) {

			//copy, don't transform
			for( k = 0; k < 4; k++ )
				poutbonepose->quat[k] = LittleFloat ( pinbonepose->quat[k] );
			for( k = 0; k < 3; k++ )
				poutbonepose->origin[k] = LittleFloat ( pinbonepose->origin[k] );

			//printf("R_SKP:bopoOrigin(j): %f %f %f\n", pinbonepose->origin[0], pinbonepose->origin[1], pinbonepose->origin[2]);
		}
	}

	//clean up
	free(buffer);
}


//=================
//Mod_LoadSkeletalModel
//=================
void SKM_ReadSKMfile(char* filename)
{
	FILE*		file;
	unsigned char	*buffer;

	int			i, j, l, k;
	int			pos=0, filesize=0;
	//----------------------------
	//IN
	dskmheader_t	*pinmodel;
	dskmmesh_t		*pinmesh;
	dskmvertex_t	*pinskmvert;
	dskmbonevert_t	*pinbonevert;
	dskmcoord_t		*pinstcoord;
	index_t			*pintris;
	unsigned int	*pinreferences;

	//----------------------------
	//OUT
	mskmodel_t		*poutmodel;
	mskmesh_t		*poutmesh;
	mskvertex_t		*poutskmvert;
	mskbonevert_t	*poutbonevert;
	vec2_t			*poutstcoord;
	index_t			*pouttris;
	unsigned int	*poutreferences;

	//----------------------------

	// open the file to get the size
	file = fopen( filename, "rb");
	if (!file) {
		printf("ReadSKM failed: can't open data file \"%s\".\n",filename);
		exit(1);
	}
	// get filesize
	pos = ftell( file );
	fseek( file, 0, SEEK_END );
	filesize = ftell( file );
	fseek( file, pos, SEEK_SET );
	fclose( file );

	// Having the size, read it
	//-------------------------------------------------------
	buffer = malloc( filesize * sizeof(unsigned char) );
	file = fopen(filename, "rb");
	fread( buffer, filesize ,sizeof(unsigned char), file);
	fclose(file);
	
	if(!buffer){
		printf("ReadSKM failed: Empty data file \"%s\".\n",filename);
		exit(1);
	}
	printf("Reading SKM model %s\n", filename);
	//-------------------------------------------------------


	poutmodel = malloc( sizeof(mskmodel_t) );
	pinmodel = ( dskmheader_t * )buffer;
	
	printf("%s: Type:%i\n", filename, LittleLong (pinmodel->type));
	if( LittleLong (pinmodel->type) != SKM_MODELTYPE ) {
		printf( "%s has wrong type number (%i should be %i)",
				 filename, LittleLong (pinmodel->type), SKM_MODELTYPE );
	}
	printf("%s: Filesize:%i\n", filename, LittleLong (pinmodel->filesize));
	if( LittleLong (pinmodel->filesize) > SKM_MAX_FILESIZE ) {
		printf( "%s has has wrong filesize (%i should be less than %i)",
				 filename, LittleLong (pinmodel->filesize), SKM_MAX_FILESIZE );
	}

	poutmodel = malloc( sizeof(mskmodel_t) );
	poutmodel->nummeshes = LittleLong( pinmodel->num_meshes );
	if( poutmodel->nummeshes <= 0 )
		printf("%s has no meshes", filename );
	else if( poutmodel->nummeshes > SKM_MAX_MESHES )
		printf("%s has too many meshes", filename );

	poutmodel->numbones = LittleLong( pinmodel->num_bones );
	if( poutmodel->numbones <= 0 )
		printf("%s has no bones", filename );
	else if( poutmodel->numbones > SKM_MAX_BONES )
		printf("%s has too many bones", filename );

	
	pinmesh = ( dskmmesh_t * )( ( unsigned char * )pinmodel + LittleLong ( pinmodel->ofs_meshes ) );
	poutmesh = poutmodel->meshes = malloc ( sizeof(mskmesh_t) * poutmodel->nummeshes );

	//------------

	for( i = 0; i < (int)poutmodel->nummeshes; i++, pinmesh++, poutmesh++ ) {
		poutmesh->numverts = LittleLong ( pinmesh->num_verts );
		if( poutmesh->numverts <= 0 )
			printf("mesh %i in model %s has no vertexes", i, filename );
		else if( poutmesh->numverts > SKM_MAX_VERTS )
			printf("mesh %i in model %s has too many vertexes", i, filename );

		poutmesh->numtris = LittleLong ( pinmesh->num_tris );
		if( poutmesh->numtris <= 0 )
			printf("mesh %i in model %s has no indices", i, filename );
		else if( poutmesh->numtris > SKM_MAX_TRIS )
			printf("mesh %i in model %s has too many indices", i, filename );

		poutmesh->numreferences = LittleLong ( pinmesh->num_references );
		if( poutmesh->numreferences <= 0 )
			printf("mesh %i in model %s has no bone references", i, filename );
		else if( poutmesh->numreferences > SKM_MAX_BONES )
			printf("mesh %i in model %s has too many bone references", i, filename );

		
		strncpy( poutmesh->name, pinmesh->meshname, sizeof(poutmesh->name) );//SKM_MAX_NAME
		//Mod_StripLODSuffix( poutmesh->name );//jalfixme

		//BY JAL: shadername is NOT stored in QF (it registers the actual shader)
		strncpy( poutmesh->shadername, pinmesh->shadername, SKM_MAX_NAME );
		//---------------------------------------------------------
		poutmesh->skin.shader = NULL;
	

		pinskmvert = ( dskmvertex_t * )( ( unsigned char * )pinmodel + LittleLong( pinmesh->ofs_verts ) );
		poutskmvert = poutmesh->vertexes = malloc( sizeof(mskvertex_t) * poutmesh->numverts );

		for( j = 0; j < (int)poutmesh->numverts; j++, poutskmvert++/*, pinskmvert++*/ ) {//JAL: pinskmvert++ by me
			poutskmvert->numbones = LittleLong( pinskmvert->numbones );

			pinbonevert = ( dskmbonevert_t * )( ( unsigned char * )pinskmvert + sizeof(poutskmvert->numbones) );
			poutbonevert = poutskmvert->verts = malloc( sizeof(mskbonevert_t) * poutskmvert->numbones );

			for( l = 0; l < (int)poutskmvert->numbones; l++, pinbonevert++, poutbonevert++ ) {
				for( k = 0; k < 3; k++ ) {
					poutbonevert->origin[k] = LittleFloat( pinbonevert->origin[k] );
					poutbonevert->normal[k] = LittleFloat( pinbonevert->normal[k] );
				}

				poutbonevert->influence = LittleFloat( pinbonevert->influence );
				poutbonevert->bonenum = LittleLong( pinbonevert->bonenum );
			}

			pinskmvert = ( dskmvertex_t * )( ( unsigned char * )pinbonevert );	//JAL: moved to pinskmvert++
		}

		pinstcoord = ( dskmcoord_t * )( ( unsigned char * )pinmodel + LittleLong (pinmesh->ofs_texcoords) );
		poutstcoord = poutmesh->stcoords = malloc( poutmesh->numverts * sizeof(vec2_t) );

		for( j = 0; j < (int)poutmesh->numverts; j++, pinstcoord++/*, poutstcoord++*/ ) {
			poutstcoord[j][0] = LittleFloat( pinstcoord->st[0] );
			poutstcoord[j][1] = LittleFloat( pinstcoord->st[1] );
		}

		pintris = ( index_t * )( ( unsigned char * )pinmodel + LittleLong (pinmesh->ofs_indices) );
		pouttris = poutmesh->indexes = malloc( sizeof(index_t) * poutmesh->numtris * 3 );

		for( j = 0; j < (int)poutmesh->numtris; j++, pintris += 3, pouttris += 3 ) {
			pouttris[0] = (index_t)LittleLong( pintris[0] );
			pouttris[1] = (index_t)LittleLong( pintris[1] );
			pouttris[2] = (index_t)LittleLong( pintris[2] );
		}

		pinreferences = ( index_t *)( ( unsigned char * )pinmodel + LittleLong (pinmesh->ofs_references) );
		poutreferences = poutmesh->references = malloc( sizeof(unsigned int) * poutmesh->numreferences );

		for( j = 0; j < (int)poutmesh->numreferences; j++, pinreferences++, poutreferences++ ) {
			*poutreferences = LittleLong( *pinreferences );
		}

	}

	//clean up
	free(buffer);

	//--------------------------------------
	//	time for loading the SKP file
	//--------------------------------------
	SKM_ReadSKPfile("data/tris.skp", poutmodel);

	modelSKM = poutmodel;

	//--------------------------------------
	//	Set up for animations
	//--------------------------------------
	SKM_InitBoneKeys(modelSKM);

	if(!LoadAnimationConfig(modelSKM))	//not yet implemented
	{
		//create a single animation covering all the frames
		CreateAnimationConfig(modelSKM);
	}
}



#define MAX_BONE_KEYS 256
static BOOL				BoneKey[MAX_BONE_KEYS];
static int				BoneKeyFrame[MAX_BONE_KEYS];
static int				totalBoneKeys;
//static 
mskbonekeys_t			printBoneKeys;

void SKM_UpdateBoneKeysRegistration(void)
{
	int		i;
	int		key;
	int		countBoneKeys = 0;
	mskmodel_t *model = modelSKM;
	
	//set all off, but preserve frame values
	for (i=0; i<MAX_BONE_KEYS; i++)
		BoneKey[i] = FALSE;

	printf("New BONE NAMES list:\n");

	//find keys in bone names
	for (i=0; i<(int)model->numbones; i++) {

		printf("%s\n",model->bones[i].name);

		key = model->bones[i].name[0];
		if (key > MAX_BONE_KEYS-1)	//not supported char
			continue;

		BoneKey[key] = TRUE;
		countBoneKeys++;
	}

	totalBoneKeys = countBoneKeys;

	SKM_SetNewFrame(modelSKM);
}

void SKM_InitBoneKeys(mskmodel_t *model)
{
	int i;
	for (i=0; i<MAX_BONE_KEYS; i++){
		BoneKey[i] = FALSE;
		BoneKeyFrame[i] = 1;	//init with 1
	}

	SKM_UpdateBoneKeysRegistration();
}
int SKM_BoneKeysAction(int key)
{
	//pressing a key

	if(!BoneKey[key])
		return FALSE;

	BoneKeyFrame[key]++;

	if(BoneKeyFrame[key] < 0)
		BoneKeyFrame[key] = (int)(modelSKM->numframes - 1);
	else if(BoneKeyFrame[key] >= (int)(modelSKM->numframes) )
		BoneKeyFrame[key] = 0;

	// update frame pose handle
	SKM_SetNewFrame(modelSKM);

	return TRUE;
}
void SKM_PrintBoneKeys(void)
{
	int	i;
	char	s[2];
	float	ypos = 0.98f;
	mskbonekeys_t *boneKeys = &printBoneKeys;

	for(i=0; i<boneKeys->numbonekeys; i++) 
	{
		s[0] = (unsigned char)boneKeys->keys[i]; s[1] = 0;
		SCRPrint (0.8f, ypos - (i*0.03f), "Key: %s frame:%i",
			s ,boneKeys->frames[i]);
	}
}
void UpdatePrintBoneKeys(mskbonekeys_t *boneKeys)
{
	//copy into printBoneKeys for printing

	if(printBoneKeys.keys)
		free(printBoneKeys.keys);
	printBoneKeys.keys = malloc(sizeof(int)*boneKeys->numbonekeys);
	if(printBoneKeys.frames)
		free(printBoneKeys.frames);
	printBoneKeys.frames = malloc(sizeof(int)*boneKeys->numbonekeys);
	printBoneKeys.numbonekeys = boneKeys->numbonekeys;
	memcpy(printBoneKeys.keys, boneKeys->keys, sizeof(int)*boneKeys->numbonekeys);
	memcpy(printBoneKeys.frames, boneKeys->frames, sizeof(int)*boneKeys->numbonekeys);
}

struct mskbonekeys_s *SKM_GetBoneKeys(void)
{
	mskbonekeys_t *boneKeys = &printBoneKeys;

	if (boneKeys->numbonekeys)
		return boneKeys;

	return NULL;
}

//----------------------------
//	Bone & Mesh Selection
//----------------------------
int	show_quats;
int	SelectedBone = 1;
int	SelectedMesh = 0;

void BoneSelectNext(int value)
{
	SelectedBone += value;
	if(SelectedBone >= (int)modelSKM->numbones)
		SelectedBone = 0;
	else if(SelectedBone < 0)
		SelectedBone = (int)(modelSKM->numbones - 1);
}
void BoneToogleAttachFlag(void)
{
	if(modelSKM->bones[SelectedBone].flags & SKM_BONEFLAG_ATTACH)
		modelSKM->bones[SelectedBone].flags &= ~SKM_BONEFLAG_ATTACH;
	else
		modelSKM->bones[SelectedBone].flags |= SKM_BONEFLAG_ATTACH;
}

void MeshSelectNext(int value)
{
	SelectedMesh += value;
	if(SelectedMesh >= (int)modelSKM->nummeshes)
		SelectedMesh = 0;
	else if(SelectedMesh < 0)
		SelectedMesh = (int)(modelSKM->nummeshes - 1);
}

//------------------------------
//	operations
//------------------------------
void Matrix_Copy( vec3_t m1[3], vec3_t m2[3] )
{
	int i, j;

	for( i = 0; i < 3; i++ )
		for( j = 0; j < 3; j++ )
			m2[i][j] = m1[i][j];
}
void Matrix_Multiply( vec3_t m1[3], vec3_t m2[3], vec3_t out[3] )
{
	out[0][0] = m1[0][0]*m2[0][0] + m1[0][1]*m2[1][0] + m1[0][2]*m2[2][0];
	out[0][1] = m1[0][0]*m2[0][1] + m1[0][1]*m2[1][1] + m1[0][2]*m2[2][1];
	out[0][2] = m1[0][0]*m2[0][2] + m1[0][1]*m2[1][2] + m1[0][2]*m2[2][2];
	out[1][0] = m1[1][0]*m2[0][0] + m1[1][1]*m2[1][0] +	m1[1][2]*m2[2][0];
	out[1][1] = m1[1][0]*m2[0][1] + m1[1][1]*m2[1][1] + m1[1][2]*m2[2][1];
	out[1][2] = m1[1][0]*m2[0][2] + m1[1][1]*m2[1][2] +	m1[1][2]*m2[2][2];
	out[2][0] = m1[2][0]*m2[0][0] + m1[2][1]*m2[1][0] +	m1[2][2]*m2[2][0];
	out[2][1] = m1[2][0]*m2[0][1] + m1[2][1]*m2[1][1] +	m1[2][2]*m2[2][1];
	out[2][2] = m1[2][0]*m2[0][2] + m1[2][1]*m2[1][2] +	m1[2][2]*m2[2][2];
}
#define DEG2RAD( a ) ( a * M_PI ) / 180.0F
void Matrix_Rotate( vec3_t m[3], vec_t angle, vec_t x, vec_t y, vec_t z )
{
	vec3_t t[3], b[3];
	vec_t c = cos( DEG2RAD(angle) );
	vec_t s = sin( DEG2RAD(angle) );
	vec_t mc = 1 - c, t1, t2;
	
	t[0][0] = (x * x * mc) + c;
	t[1][1] = (y * y * mc) + c;
	t[2][2] = (z * z * mc) + c;

	t1 = y * x * mc;
	t2 = z * s;
	t[0][1] = t1 + t2;
	t[1][0] = t1 - t2;

	t1 = x * z * mc;
	t2 = y * s;
	t[0][2] = t1 - t2;
	t[2][0] = t1 + t2;

	t1 = y * z * mc;
	t2 = x * s;
	t[1][2] = t1 + t2;
	t[2][1] = t1 - t2;

	Matrix_Copy( m, b );
	Matrix_Multiply( b, t, m );
}

//----------------------------------------------------
vec3_t		vec3_origin = { 0, 0, 0 };
vec3_t		axis_identity[3] = { { 1, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };
//----------------------------------------------------

void SKM_RotateSkeleton(void)
{
	int		j, k;
	float	angle;
	mskmodel_t *model;
	vec3_t		axis_rotator[3];
	quat_t		quat_rotator;

	model = modelSKM;

	angle = -90;

	Matrix_Copy( axis_identity, axis_rotator );
	Matrix_Rotate( axis_rotator, angle, 0.0f, 0.0f, 1.0f );
	Matrix_Quat( axis_rotator, quat_rotator );

	//start from frame 1, ignore pose frame
	for( j = 1; j < (int)model->numframes; j++ ) 
	{	
		//project bip0 to quat_rotator at every frame
		mskbonepose_t	temppose;	//helper
		
		for( k = 0; k < 4; k++ )
			temppose.quat[k] = model->frames[j].boneposes[0].quat[k];
		for( k = 0; k < 3; k++ )
			temppose.origin[k] = model->frames[j].boneposes[0].origin[k];
		
		Quat_ConcatTransforms ( quat_rotator, vec3_origin, temppose.quat, temppose.origin, model->frames[j].boneposes[0].quat, model->frames[j].boneposes[0].origin );
		
	}

	SKM_SetNewFrame(modelSKM);
}

void SKM_Rotate(float angles[3])
{
	
	int		j, k;
	//float	angle;
	mskmodel_t *model;
	vec3_t		axis_rotator[3];
	quat_t		quat_rotator;

	model = modelSKM;

	//angle = -90;

	Matrix_Copy( axis_identity, axis_rotator );
	Matrix_Rotate( axis_rotator, angles[YAW], 0.0f, 0.0f, 1.0f );
	Matrix_Rotate( axis_rotator, angles[PITCH], 0.0f, 1.0f, 0.0f );
	Matrix_Rotate( axis_rotator, angles[ROLL], 1.0f, 0.0f, 0.0f );
	Matrix_Quat( axis_rotator, quat_rotator );

	//start from frame 1, ignore pose frame
	for( j = 1; j < (int)model->numframes; j++ ) 
	{	
		//project bip0 to quat_rotator at every frame
		mskbonepose_t	temppose;	//helper
		
		for( k = 0; k < 4; k++ )
			temppose.quat[k] = model->frames[j].boneposes[0].quat[k];
		for( k = 0; k < 3; k++ )
			temppose.origin[k] = model->frames[j].boneposes[0].origin[k];
		
		Quat_ConcatTransforms ( quat_rotator, vec3_origin, temppose.quat, temppose.origin, model->frames[j].boneposes[0].quat, model->frames[j].boneposes[0].origin );
		
	}

	SKM_SetNewFrame(modelSKM);

}

//==============================
//frame pose pair
//typedef struct msk_framepose_s
//{
//	mskbonepose_t	*boneposes;
//	mskbonepose_t	*oldboneposes;
//
//} msk_frameposes_t;
//==============================
static msk_frameposes_t	framePose;
//==============================

//=======================
//	SKM_Skel_NewFramePose
//	create a new pose with the info provided by bonekeys
//=======================
void SKM_Skel_NewFramePose(mskmodel_t *model, msk_frameposes_t *framepose, mskbonekeys_t *boneKeys)
{
	int				j, k, i;
	mskbonepose_t	temppose;	//helper
	mskframe_t		*uframe;	//Get untransformed frame


	//free old
	if(framepose->oldboneposes)
		free(framepose->oldboneposes);

	//point oldframe to newframe
	framepose->oldboneposes = framepose->boneposes;
	//alloc a new newframe
	framepose->boneposes = malloc ( sizeof(mskbonepose_t) * model->numbones );
	//if still there's nothing in oldframe it means there wasn't originally
	//nothing in new. So point both to new, cause it's our first frame
	if(framepose->oldboneposes)
		framepose->oldboneposes = framepose->boneposes;

	

	//create/transform into the newframe pose
	for( j = 0; j < (int)model->numbones; j++ ) 
	{
		//find this bone frame handle
		uframe = &model->frames[0];//frame zero for unrecognized BoneKey characters
		for( i=0; i<boneKeys->numbonekeys; i++) 
		{
			if(model->bones[j].name[0] == boneKeys->keys[i])
				uframe = &model->frames[boneKeys->frames[i]];
		}	
		
		if( model->bones[j].parent >= 0 )
		{
			for( k = 0; k < 4; k++ )
				temppose.quat[k] = uframe->boneposes[j].quat[k];
			for( k = 0; k < 3; k++ )
				temppose.origin[k] = uframe->boneposes[j].origin[k];
			
			Quat_ConcatTransforms ( framepose->boneposes[model->bones[j].parent].quat, framepose->boneposes[model->bones[j].parent].origin, temppose.quat, temppose.origin, framepose->boneposes[j].quat, framepose->boneposes[j].origin );
			
		} else {
			for( k = 0; k < 4; k++ )
				framepose->boneposes[j].quat[k] = uframe->boneposes[j].quat[k];
			for( k = 0; k < 3; k++ )
				framepose->boneposes[j].origin[k] = uframe->boneposes[j].origin[k];
		}	
	}
}

//=======================
//	SKM_SetNewFrame
//	create a mskbonekeys_t struct with the keys needed to update the framepose.
//=======================
void SKM_SetNewFrame(mskmodel_t *model)
{
	int		i;
	mskbonekeys_t boneKeys;

	//make the keys & frames list
	boneKeys.numbonekeys = 0;
	boneKeys.keys = malloc(sizeof(int)*totalBoneKeys);
	boneKeys.frames = malloc(sizeof(int)*totalBoneKeys);
	//link them into the list
	for ( i=0; i<256; i++ ) {
		if(BoneKey[i] == TRUE) {
			boneKeys.keys[boneKeys.numbonekeys] = i;
			boneKeys.frames[boneKeys.numbonekeys] = BoneKeyFrame[i];
			boneKeys.numbonekeys++;
		}
	}

	//copy into printBoneKeys for printing
	UpdatePrintBoneKeys(&boneKeys);

	//create skeleton framepose
	SKM_Skel_NewFramePose(modelSKM, &framePose, &boneKeys);
	//free
	free(boneKeys.keys);
	free(boneKeys.frames);
}

//=======================
//	SKMDrawTransformSkeleton
//	draw model's skeleton
//=======================
void SKM_DrawSkeleton(mskmodel_t *model, msk_frameposes_t *framepose)
{
	int				j;
	vec3_t			color, colorTag, colorSelected, vecZero;

	//colors
	color[RED] = 0.4f;color[GREEN] = 0.5f;color[BLUE] = 0.7f;
	colorSelected[RED] = 1.0f;colorSelected[GREEN] = 0.0f;colorSelected[BLUE] = 0.0f;
	colorTag[RED] = 1.0f;colorTag[GREEN] = 1.0f;colorTag[BLUE] = 1.0f;
	vecZero[0] = 0.0f;vecZero[1] = 0.0f;vecZero[2] = 0.0f;


	//print active keys
	SKM_PrintBoneKeys();

	
	//draw external new frame pose
	for( j = 0; j < (int)model->numbones; j++)
	{
		GL_AddCubeToSceneAtLoc(framepose->boneposes[j].origin, color, vecZero, 0.5f);
		if (model->bones[j].parent >= 0) 
		{
			if( j == SelectedBone)
				GL_AddLineToScene( framepose->boneposes[model->bones[j].parent].origin, framepose->boneposes[j].origin, colorSelected );
			else if(model->bones[j].flags & SKM_BONEFLAG_ATTACH)
				GL_AddLineToScene( framepose->boneposes[model->bones[j].parent].origin, framepose->boneposes[j].origin, colorTag );
			else
				GL_AddLineToScene( framepose->boneposes[model->bones[j].parent].origin, framepose->boneposes[j].origin, color );

			if(show_quats) {
				
				vec3_t	tmp[3], tmp1, tmp2, tmp3, colorQuat;
				colorQuat[RED] = 0.0f;colorQuat[GREEN] = 1.0f;colorQuat[BLUE] = 0.0f;

				Quat_Matrix( framepose->boneposes[model->bones[j].parent].quat, tmp );

				tmp1[0] = tmp[0][0] + framepose->boneposes[model->bones[j].parent].origin[0];
				tmp1[1] = tmp[1][0] + framepose->boneposes[model->bones[j].parent].origin[1];
				tmp1[2] = tmp[2][0] + framepose->boneposes[model->bones[j].parent].origin[2];

				tmp2[0] = tmp[0][1] + framepose->boneposes[model->bones[j].parent].origin[0];
				tmp2[1] = tmp[1][1] + framepose->boneposes[model->bones[j].parent].origin[1];
				tmp2[2] = tmp[2][1] + framepose->boneposes[model->bones[j].parent].origin[2];

				tmp3[0] = tmp[0][2] + framepose->boneposes[model->bones[j].parent].origin[0];
				tmp3[1] = tmp[1][2] + framepose->boneposes[model->bones[j].parent].origin[1];
				tmp3[2] = tmp[2][2] + framepose->boneposes[model->bones[j].parent].origin[2];

				GL_AddLineToScene( framepose->boneposes[model->bones[j].parent].origin, tmp1, colorQuat );
				GL_AddLineToScene( framepose->boneposes[model->bones[j].parent].origin, tmp2, colorQuat );
				GL_AddLineToScene( framepose->boneposes[model->bones[j].parent].origin, tmp3, colorQuat );
				
			}
		}
	}
}


void SKM_AnimToFrame (mskmodel_t *model, animationinfo_t *anim, mskbonekeys_t *boneKeys)
{

	int i;
	//animationconfig_t *pmodelinfo = &anim->config;

	//splitskel versioned

	//interpolate
	if (timer.time < anim->nextframetime)
	{
		anim->backlerp = 1.0f - ((timer.time - anim->prevframetime)/(anim->nextframetime - anim->prevframetime));
		if (anim->backlerp > 1)
			anim->backlerp = 1;
		else if (anim->backlerp < 0)
			anim->backlerp = 0;
		
		return;
	}
	
	for (i=0; i<PMODEL_PARTS; i++)
	{
		//advance frames
		anim->oldframe[i] = anim->frame[i];
		anim->frame[i]++;
		
		//looping
		//if (anim->frame[i] > pmodelinfo->lastframe[anim->current[i]])
		if (anim->frame[i] > anim->config.anims[anim->current[i]].lastframe)
		{
			if (anim->currentChannel[i])	
				anim->currentChannel[i] = BASIC_CHANNEL;//kill

			//anim->frame[i] = (pmodelinfo->lastframe[anim->current[i]] - pmodelinfo->loopingframes[anim->current[i]]);
			anim->frame[i] = (anim->config.anims[anim->current[i]].lastframe - anim->config.anims[anim->current[i]].loopingframes);
		}

		//NEWANIM
		if ( anim->buffer[EVENT_CHANNEL].newanim[i] )
		{
			//backup if basic
			if (anim->buffer[BASIC_CHANNEL].newanim[i] +
				anim->currentChannel[i] == BASIC_CHANNEL){//both are zero
				anim->buffer[BASIC_CHANNEL].newanim[i] = anim->current[i];
			}
			//set up
			anim->current[i] = anim->buffer[EVENT_CHANNEL].newanim[i];
			anim->frame[i] = anim->config.anims[anim->current[i]].firstframe;
			anim->currentChannel[i] = EVENT_CHANNEL;
			anim->buffer[EVENT_CHANNEL].newanim[i] = 0;
		}
		else if ( anim->buffer[BASIC_CHANNEL].newanim[i]
			&& ( anim->currentChannel[i] != EVENT_CHANNEL ) )
		{
			//set up
			anim->current[i] = anim->buffer[BASIC_CHANNEL].newanim[i];
			anim->frame[i] = anim->config.anims[anim->current[i]].firstframe;
			anim->currentChannel[i] = BASIC_CHANNEL;
			anim->buffer[BASIC_CHANNEL].newanim[i] = 0;
		}
	}

	//build new frameposes
//	CG_PModelSKM_NewFramePose(pmodel);//splitskelmain
	//updated frametime
	anim->prevframetime = timer.time;
	anim->nextframetime = timer.time + anim->config.frametime;
	anim->backlerp = 1.0f;
	
}

void SKMDrawModel(void)
{
	SKM_AnimToFrame( modelSKM, &skmanim, &printBoneKeys );
	SKM_DrawSkeleton( modelSKM, &framePose );
}



// tool management
//----------------------------------------------------------
void ChangeBoneName( void )
{
	mskmodel_t		*model;
	model = modelSKM;	//tmp

	InputAction( "Introduce New Bone Name: (ESC to cancel)", model->bones[SelectedBone].name, SKM_MAX_NAME, SKM_UpdateBoneKeysRegistration);
}

void SKM_DrawModelInfo(void)
{
	mskmodel_t		*model;
	model = modelSKM;	//tmp
	SCRPrint ( 0.02f, 0.09f, "Model: %s", "data/tris");//jalfixme
	SCRPrint ( 0.02f, 0.06f, "Selected Bone: %s", model->bones[SelectedBone].name);
}
//----------------------------------------------------------



void Change_SelectedBoneName( char *newname )
{
	mskmodel_t		*model;
	model = modelSKM;	//tmp

	if(!newname)
		return;

	if(!strcmp(model->bones[SelectedBone].name, newname))
		return;

	strncpy(model->bones[SelectedBone].name, newname, SKM_MAX_NAME);

	SKM_UpdateBoneKeysRegistration();
}

char *getSelectedBoneName(void)
{
	mskmodel_t		*model;
	model = modelSKM;	//tmp
	

	//if(SelectedBone < 0 || !model->bones[SelectedBone].name[0])
	//	return NULL;

	return model->bones[SelectedBone].name;
}

void Change_SelectedMeshName( char *newname )
{
	mskmodel_t		*model;
	model = modelSKM;	//tmp

	if(!newname)
		return;

	if(!strcmp(model->meshes[SelectedMesh].name, newname))
		return;

	strncpy(model->meshes[SelectedMesh].name, newname, SKM_MAX_NAME);
}

char *getSelectedMeshName(void)
{
	mskmodel_t		*model;
	model = modelSKM;	//tmp
	

	//if(SelectedBone < 0 || !model->bones[SelectedBone].name[0])
	//	return NULL;

	return model->meshes[SelectedMesh].name;
}


void Change_SelectedSkinMeshName( char *newname )
{
	mskmodel_t		*model;
	model = modelSKM;	//tmp

	if(!newname)
		return;

	if(!strcmp(model->meshes[SelectedMesh].shadername, newname))
		return;

	strncpy(model->meshes[SelectedMesh].shadername, newname, SKM_MAX_NAME);
}

char *getSelectedSkinMeshName(void)
{
	mskmodel_t		*model;
	model = modelSKM;	//tmp
	

	//if(SelectedBone < 0 || !model->bones[SelectedBone].name[0])
	//	return NULL;

	return model->meshes[SelectedMesh].shadername;
}
